﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Analysis.AlgorithmBehavior.Analyzers;
using HeuristicLab.Core.Views;
using HeuristicLab.Data;
using HeuristicLab.MainForm;
using HeuristicLab.Optimization;

namespace HeuristicLab.Analysis.AlgorithmBehavior.Views {
  [View("Statistical Testing View")]
  [Content(typeof(RunCollection), false)]
  public sealed partial class StatisticalTestingView : ItemView {
    private double[][] data;

    public StatisticalTestingView() {
      InitializeComponent();
    }

    public new RunCollection Content {
      get { return (RunCollection)base.Content; }
      set { base.Content = value; }
    }

    public override bool ReadOnly {
      get { return true; }
      set { /*not needed because results are always readonly */}
    }

    protected override void OnContentChanged() {
      base.OnContentChanged();

      if (Content != null) {
        UpdateResultComboBox();
        UpdateGroupsComboBox();
        RebuildDataTable();
      }
    }

    #region events
    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
    }

    protected override void DeregisterContentEvents() {
      base.DeregisterContentEvents();
    }
    #endregion

    private void UpdateGroupsComboBox() {
      groupComboBox.Items.Clear();

      var parameters = (from run in Content
                        from param in run.Parameters
                        select param.Key).Distinct().ToArray();

      foreach (var p in parameters) {
        var variations = (from run in Content
                          where run.Parameters.ContainsKey(p) &&
                          (run.Parameters[p] is IntValue || run.Parameters[p] is DoubleValue ||
                          run.Parameters[p] is StringValue || run.Parameters[p] is BoolValue)
                          select ((dynamic)run.Parameters[p]).Value).Distinct();

        if (variations.Count() > 1) {
          groupComboBox.Items.Add(p);
        }
      }
      if (groupComboBox.Items.Count > 0) groupComboBox.SelectedItem = groupComboBox.Items[0];
    }

    private string[] GetColumnNames(IEnumerable<IRun> runs) {
      string parameterName = (string)groupComboBox.SelectedItem;
      var r = runs.Where(x => x.Parameters.ContainsKey(parameterName));
      return r.Select(x => ((dynamic)x.Parameters[parameterName]).Value).Distinct().Select(x => (string)x.ToString()).ToArray();
    }

    private void UpdateResultComboBox() {
      resultComboBox.Items.Clear();
      var results = (from run in Content
                     from result in run.Results
                     where result.Value is IntValue || result.Value is DoubleValue
                     select result.Key).Distinct().ToArray();

      resultComboBox.Items.AddRange(results);
      if (resultComboBox.Items.Count > 0) resultComboBox.SelectedItem = resultComboBox.Items[0];
    }

    private void RebuildDataTable() {
      string parameterName = (string)groupComboBox.SelectedItem;
      if (parameterName != null) {
        string resultName = (string)resultComboBox.SelectedItem;

        var runs = Content.Where(x => x.Results.ContainsKey(resultName) && x.Visible);
        var columnNames = GetColumnNames(runs);
        var groups = GetGroups(columnNames, runs);
        data = new double[columnNames.Count()][];

        DoubleMatrix dt = new DoubleMatrix(groups.Select(x => x.Count()).Max(), columnNames.Count());
        dt.ColumnNames = columnNames;

        int i = 0;
        int j = 0;
        foreach (string columnName in columnNames) {
          j = 0;
          data[i] = new double[groups[i].Count()];
          foreach (IRun run in groups[i]) {
            dt[j, i] = (double)((dynamic)run.Results[resultName]).Value;
            data[i][j] = dt[j, i];
            j++;
          }
          i++;
        }

        stringConvertibleMatrixView.Content = dt;
      }
    }

    private List<RunCollection> GetGroups(string[] columnNames, IEnumerable<IRun> runs) {
      List<RunCollection> runCols = new List<RunCollection>();
      string parameterName = (string)groupComboBox.SelectedItem;

      foreach (string cn in columnNames) {
        var tmpRuns = runs.Where(x => ((string)((dynamic)x.Parameters[parameterName]).Value.ToString()) == cn);
        runCols.Add(new RunCollection(tmpRuns));
      }

      return runCols;
    }

    private void testButton_Click(object sender, EventArgs e) {
      double pval = KruskalWallis.Test(data);
      pValTextBox.Text = pval.ToString();
    }

    private void normalDistButton_Click(object sender, EventArgs e) {
      double val;
      List<double> res = new List<double>();

      for (int i = 0; i < data.Length; i++) {
        alglib.jarqueberatest(data[i], data[i].Length, out val);
        res.Add(val);
      }

      for (int i = 0; i < res.Count(); i++) {
        if (res[i] < 0.1) {
          stringConvertibleMatrixView.DataGridView.Columns[i].DefaultCellStyle.ForeColor = Color.Red;
        } else {
          stringConvertibleMatrixView.DataGridView.Columns[i].DefaultCellStyle.ForeColor = Color.Green;
        }
      }
    }

    private void resultComboBox_SelectedValueChanged(object sender, EventArgs e) {
      RebuildDataTable();
      pValTextBox.Text = string.Empty;
    }

    private void groupComboBox_SelectedValueChanged(object sender, EventArgs e) {
      RebuildDataTable();
      pValTextBox.Text = string.Empty;
    }

    private void normalityDetails_Click(object sender, EventArgs e) {
      DoubleMatrix pValsMatrix = new DoubleMatrix(1, stringConvertibleMatrixView.Content.Columns);
      pValsMatrix.ColumnNames = stringConvertibleMatrixView.Content.ColumnNames;
      pValsMatrix.RowNames = new string[] { "p-Values" };

      double val;
      for (int i = 0; i < data.Length; i++) {
        alglib.jarqueberatest(data[i], data[i].Length, out val);
        pValsMatrix[0, i] = val;
      }

      MainFormManager.MainForm.ShowContent(pValsMatrix);
    }

    private void pairwiseTestButton_Click(object sender, EventArgs e) {
      var selectedCells = stringConvertibleMatrixView.DataGridView.SelectedCells;
      if (selectedCells.Count < 1) {
        MessageBox.Show("Please selected one cell/column for pairwise comparision. ", "HeuristicLab", MessageBoxButtons.OK, MessageBoxIcon.Error);
        return;
      }

      int colIndex = selectedCells[0].ColumnIndex;
      foreach (DataGridViewCell selC in selectedCells) {
        if (colIndex != selC.ColumnIndex) {
          MessageBox.Show("Please selected only one column for pairwise comparision. ", "HeuristicLab", MessageBoxButtons.OK, MessageBoxIcon.Error);
          return;
        }
      }

      DoubleMatrix pValsMatrix = new DoubleMatrix(1, stringConvertibleMatrixView.Content.Columns);
      pValsMatrix.ColumnNames = stringConvertibleMatrixView.Content.ColumnNames;
      pValsMatrix.RowNames = new string[] { "p-Values" };

      double bothtails;
      double lefttail;
      double righttail;
      for (int i = 0; i < data.Length; i++) {
        alglib.mannwhitneyutest(data[colIndex], data[colIndex].Length, data[i], data[i].Length, out bothtails, out lefttail, out righttail);
        pValsMatrix[0, i] = bothtails;
      }

      MainFormManager.MainForm.ShowContent(pValsMatrix);
    }

    private void infoLabel_DoubleClick(object sender, EventArgs e) {
      using (TextDialog dialog = new TextDialog("Description of statistical tests", toolTip1.GetToolTip(this.infoLabel), true)) {
        dialog.ShowDialog(this);
      }
    }
  }
}
